import hikari
import lightbulb

__all__ = [
    "Option",
    "Params"
]

class Option():
    def __init__(self, name : str, description : str, var_type : hikari.OptionType) -> None:
        self.name = name
        self.description = description
        self.var_type = var_type

    def process_module(self, mod):
        mod.command = lightbulb.option(self.name, self.description, self.var_type)(mod.command)

class Params():
    def __init__(self, description : str, options=[], name="") -> None:
        self.description = description
        self.options = options
        self.name = name

    def process_module(self, mod):
        mod.command = lightbulb.command(self.name, self.description)(mod.command)
        for opt in self.options:
            opt.process_module(mod)



import lightbulb
import hikari
import os
import glob
import logging
from .params import *
from .errors import *
from .properties import *
from importlib.util import spec_from_file_location

__all__ = [
    "Plugin",
    "CommandContext"
]

_LOGGER = logging.getLogger("hecate.plugin")

class CommandContext():
    def __init__(self, category: lightbulb.commands.Command, params: Params, properties: Properties, path: str) -> None:
        self.category = category
        self.params = params
        self.properties = properties
        self.path = path

class Plugin(lightbulb.Plugin):
    def __init__(self, name: str, extension_path: str, default_properties={}, on_command_error=None, on_event_error=None, on_method_enter=None, on_method_exit=None) -> None:
        super().__init__(name)

        self.__command_types = {
            "prefix-commands": lightbulb.commands.PrefixCommand,
            "message-commands": lightbulb.commands.MessageCommand,
            "user-commands": lightbulb.commands.UserCommand,
            "slash-commands": lightbulb.commands.SlashCommand
        }

        self.__properties = default_properties

        def enter_exit_decorator(com_ctx):
            def dec(func):
                async def new_func(ctx):
                    if on_method_enter:
                        await on_method_enter(com_ctx, ctx)
                    await func(ctx)
                    if on_method_exit:
                        await on_method_exit(com_ctx, ctx)
                return new_func    
            return dec            

        def catch_command_exceptions(func):
            async def new_func(ctx):
                try:
                    await func(ctx)
                except:
                    await on_command_error(ctx)
            return new_func

        def catch_event_exceptions(func):
            async def new_func(e):
                try:
                    await func(e)
                except:
                    await on_event_error(e)
            return new_func

        dir_path = os.path.dirname(extension_path)
        for command in self.__command_types:
            grab_path = os.path.join(dir_path, command)
            if not os.path.exists(grab_path):
                continue
            
            py_valid = 0
            for py_file in glob.glob(os.path.join(grab_path, "*.py")):
                py_name = os.path.basename(py_file)
                if py_name.startswith('_'):
                    continue

                spec = spec_from_file_location(py_name, py_file)
                mod = spec.loader.load_module()

                if not hasattr(mod, 'command'):
                    raise MissingMethodError(f"Command module '{py_name}' doesn't declare a 'command' method")

                if hasattr(mod, 'properties') and isinstance(mod.properties, Properties):
                    mod.properties.update(self.__properties)

                params = None
                if hasattr(mod, 'params') and isinstance(mod.params, Params):
                    params = mod.params
                elif hasattr(mod, 'description') and isinstance(mod.description, str):
                    params = Params(
                        description=mod.description,
                        options=mod.options if hasattr(mod, 'options') else [],
                        name=mod.name if hasattr(mod, 'name') else py_name[:-3],
                    )
                else:
                    raise MissingParamsError(f"Command module '{py_name}' doesn't contain the necessary attributes")

                if on_method_enter != None or on_method_exit != None:
                    mod.command = enter_exit_decorator(CommandContext(
                        self.__command_types[command], 
                        params,
                        mod.properties if hasattr(mod, 'properties') else None,
                        py_file))(mod.command)
                if on_command_error != None:
                    mod.command = catch_command_exceptions(mod.command)
                mod.command = lightbulb.implements(self.__command_types[command])(mod.command)
                params.process_module(mod)
                mod.command = self.command()(mod.command)
                py_valid += 1
            if py_valid:
                _LOGGER.info(f"Fetched {py_valid} commands from '{command}'")
        
        event_path = os.path.join(dir_path, "events")
        if os.path.exists(event_path):
            py_valid = 0
            for py_file in glob.glob(os.path.join(event_path, "*.py")):
                py_name = os.path.basename(py_file)
                if py_name.startswith('_'):
                    continue

                spec = spec_from_file_location(py_name, py_file)
                mod = spec.loader.load_module()

                if not hasattr(mod, 'event'):
                    raise MissingMethodError(f"Event module '{py_name}' doesn't declare an 'event' method")

                if not hasattr(hikari, py_name[:-3]):
                    raise MissingHikariEventError(f"Event module '{py_name}' is not named after a valid event")

                if on_event_error != None:
                    mod.event = catch_event_exceptions(mod.event)
                mod.event = self.listener(getattr(hikari, py_name[:-3]))(mod.event)
                py_valid += 1
            if py_valid:
                _LOGGER.info(f"Fetched {py_valid} events")
     


from lightbulb import LightbulbError

__all__ = [
    "HecateError",
    "MissingParamsError",
    "MissingMethodError",
    "MissingHikariEventError"
]

class HecateError(LightbulbError):
    """
    Base hecate exception class. All errors raised by hecate will be a subclass
    of this exception.
    """

class MissingParamsError(HecateError):
    '''
    Exception raised when a command module accessed by a hecate `Plugin` does not
    contain a `params` attribute or the required params attributes in separate.
    '''

class MissingMethodError(HecateError):
    '''
    Exception raised when a module accessed by a hecate `Plugin` does not declare a
    `command` or `event` method.
    '''

class MissingHikariEventError(HecateError):
    '''
    Exception raised when an event module accessed by a hecate `Plugin` has a name
    that doesn't match any hikari events.
    '''
   
class PropertyBuildError(HecateError):
    '''
    Exception raised when a property object is created with a truthy value.
    '''



__version__ = "0.1.4"

__all__ = [
    "Plugin",
    "CommandContext",
    "Option",
    "Params",
    "Properties",
    "HecateError",
    "MissingMethodError",
    "MissingParamsError",
    "MissingHikariEventError",
    "PropertyBuildError"
]

from .properties import Properties
from .plugin import Plugin, CommandContext
from .params import Params, Option
from .errors import *


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['hecate']

package_data = \
{'': ['*']}

install_requires = \
['hikari-lightbulb>=2.3.1,<3.0.0']

setup_kwargs = {
    'name': 'lightbulb-ext-hecate',
    'version': '0.1.4',
    'description': 'File driven wrappers around lightbulb extensions and plugins.',
    'long_description': '# lightbulb-ext-hecate\nFile driven wrappers around lightbulb extensions and plugins.\n',
    'author': 'Radkii',
    'author_email': 'real.radkii@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/radkii/lightbulb-ext-hecate',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.12',
}


setup(**setup_kwargs)



from .errors import PropertyBuildError

__all__ = [
    "Properties"
]

class Properties():
    def __init__(self, **kwargs) -> None:
        for key in kwargs:
            if bool(kwargs[key]):
                raise PropertyBuildError(f"Property '{key}' must be initialized with an empty value.")
        self.__property_request = kwargs

    def __contains__(self, __o: object) -> bool:
        return __o in self.__property_request

    def update(self, ext_properties: dict):
        for key in self.__property_request:
            setattr(self, key, ext_properties[key] if key in ext_properties else self.property_request[key])